/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) B.A.T.M.A.N. contributors:
 *
 * Andreas Langer <an.langer@gmx.de>, Marek Lindner <marek.lindner@mailbox.org>
 *
 * License-Filename: LICENSES/preferred/GPL-2.0
 */

#ifndef _BATCTL_BATADV_PACKET_COMPAT_H
#define _BATCTL_BATADV_PACKET_COMPAT_H

#include <linux/stddef.h>

#ifndef __counted_by_be
#define __counted_by_be(m)
#endif

#include "batadv_packet.h"

#endif
